from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Any, Dict, List, Tuple, Union

import pandas as pd


class _Model(ABC):
    "Base class for all models (detectors, transformers, and aggregators)."

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def get_params(self) -> Dict[str, Any]:
        """Get the parameters of this model.

        Returns
        -------
        dict
            Model parameters.

        """
        return {key: getattr(self, key) for key in self._param_names}

    def set_params(self, **params: Any) -> None:
        """Set the parameters of this model.

        Parameters
        ----------
        **params
            Model parameters to set.

        """
        for key in params.keys():
            if key not in self._param_names:
                raise KeyError(
                    "'{}' is not a valid parameter name.".format(key)
                )
        for key, value in params.items():
            setattr(self, key, value)

    @property
    @abstractmethod
    def _param_names(self) -> Tuple[str, ...]:
        return tuple()


class _NonTrainableModel(_Model):
    "Base class of models that do not need training."

    @abstractmethod
    def _predict(self, input: Any) -> Any:
        pass

    @abstractmethod
    def _predict_core(self, input: Any) -> Any:
        pass

    @abstractmethod
    def predict(self, input: Any) -> Any:
        pass


class _TrainableModel(_Model):
    "Base class of models that need training."

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        # 0 for not fitted, 1 for fitted, 2 for univariate model fitted by DF
        self._fitted = 0  # type: int

    @abstractmethod
    def _fit(self, input: Any) -> None:
        pass

    @abstractmethod
    def _fit_core(self, input: Any) -> None:
        pass

    @abstractmethod
    def fit(self, input: Any) -> None:
        pass

    @abstractmethod
    def _predict(self, input: Any) -> Any:
        pass

    @abstractmethod
    def _predict_core(self, input: Any) -> Any:
        pass

    @abstractmethod
    def predict(self, input: Any) -> Any:
        pass

    @abstractmethod
    def fit_predict(self, input: Any) -> Any:
        pass


class _NonTrainableUnivariateModel(_NonTrainableModel):
    "Base class of univariate detectors and transformers."

    def _predict(
        self, ts: Union[pd.Series, pd.DataFrame]
    ) -> Union[pd.Series, pd.DataFrame]:
        if isinstance(ts, pd.Series):
            s = ts.copy()  # type: pd.Series
            if not isinstance(s.index, pd.DatetimeIndex):
                raise TypeError(
                    "Index of the input time series must be a pandas "
                    "DatetimeIndex object."
                )
            predicted = self._predict_core(s)
            # if a Series-to-Series operation, make sure Series name keeps
            if isinstance(predicted, pd.Series):
                predicted.name = ts.name
        elif isinstance(ts, pd.DataFrame):
            df = ts.copy()  # type: pd.DataFrame
            if df.columns.duplicated().any():
                raise ValueError(
                    "Input DataFrame must have unique column names."
                )
            # apply the model to each column
            predicted_all_cols = []
            for col in df.columns:
                predicted_this_col = self._predict(df[col])
                # if a Series-to-DF operation, update column name
                if isinstance(predicted_this_col, pd.DataFrame):
                    predicted_this_col = predicted_this_col.rename(
                        columns={
                            col1: "{}_{}".format(col, col1)
                            for col1 in predicted_this_col.columns
                        }
                    )
                predicted_all_cols.append(predicted_this_col)
            predicted = pd.concat(predicted_all_cols, axis=1)
        else:
            raise TypeError("Input must be a pandas Series or DataFrame.")
        # make sure index freq is the same (because pandas has a bug that some
        # operation, e.g. concat, may change freq)
        predicted.index.freq = ts.index.freq
        return predicted


class _TrainableUnivariateModel(_TrainableModel):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._models = dict()  # type: Dict[str, _TrainableUnivariateModel]

    def _fit(self, ts: Union[pd.Series, pd.DataFrame]) -> None:
        if isinstance(ts, pd.Series):
            s = ts.copy()  # type: pd.Series
            self._fit_core(s)
            self._fitted = 1
        elif isinstance(ts, pd.DataFrame):
            df = ts.copy()
            if not isinstance(df.index, pd.DatetimeIndex):
                raise TypeError(
                    "Index of the input time series must be a pandas "
                    "DatetimeIndex object."
                )
            if df.columns.duplicated().any():
                raise ValueError(
                    "Input DataFrame must have unique column names."
                )
            # create internal models
            self._models = {
                col: self.__class__(**deepcopy(self.get_params()))
                for col in df.columns
            }
            # fit model for each column
            for col in df.columns:
                self._models[col].fit(df[col])
            self._fitted = 2
        else:
            raise TypeError("Input must be a pandas Series or DataFrame.")

    def _predict(
        self, ts: Union[pd.Series, pd.DataFrame]
    ) -> Union[pd.Series, pd.DataFrame]:
        if self._fitted == 0:
            raise RuntimeError("The model must be trained first.")

        if isinstance(ts, pd.Series):
            if self._fitted == 2:
                raise RuntimeError(
                    "The model was trained by a pandas DataFrame object, "
                    "it can only be applied to a pandas DataFrame object with "
                    "the same column names as the one used for training."
                )
            s = ts.copy()
            if not isinstance(s.index, pd.DatetimeIndex):
                raise TypeError(
                    "Index of the input time series must be a pandas "
                    "DatetimeIndex object."
                )
            predicted = self._predict_core(s)
            # if a Series-to-Series operation, make sure Series name keeps
            if isinstance(predicted, pd.Series):
                predicted.name = ts.name
        elif isinstance(ts, pd.DataFrame):
            df = ts.copy()
            if not isinstance(df.index, pd.DatetimeIndex):
                raise TypeError(
                    "Index of the input time series must be a pandas "
                    "DatetimeIndex object."
                )
            if df.columns.duplicated().any():
                raise ValueError(
                    "Input DataFrame must have unique column names."
                )
            if self._fitted == 1:
                # apply the model to each column
                predicted_all_cols = []
                for col in df.columns:
                    predicted_this_col = self._predict(df[col])
                    if isinstance(predicted_this_col, pd.DataFrame):
                        predicted_this_col = predicted_this_col.rename(
                            columns={
                                col1: "{}_{}".format(col, col1)
                                for col1 in predicted_this_col.columns
                            }
                        )
                    predicted_all_cols.append(predicted_this_col)
                predicted = pd.concat(predicted_all_cols, axis=1)
            else:
                # predict for each column
                if not (set(self._models.keys()) >= set(df.columns)):
                    raise ValueError(
                        "The model was trained by a pandas DataFrame with "
                        "columns {}, but the input DataFrame contains columns "
                        "{} which are unknown to the model.".format(
                            list(set(self._models.keys())),
                            list(set(df.columns) - set(self._models.keys())),
                        )
                    )
                predicted = pd.concat(
                    [
                        self._models[col]._predict(df[col])
                        for col in df.columns
                    ],
                    axis=1,
                )
        else:
            raise TypeError("Input must be a pandas Series or DataFrame.")
        # make sure index freq is the same (because pandas has a bug that some
        # operation, e.g. concat, may change freq)
        predicted.index.freq = ts.index.freq
        return predicted


class _NonTrainableMultivariateModel(_NonTrainableModel):
    def _predict(self, df: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        if isinstance(df, pd.DataFrame):
            if df.columns.duplicated().any():
                raise ValueError(
                    "Input DataFrame must have unique column names."
                )
            df_copy = df.copy()
            predicted = self._predict_core(df_copy)
        else:
            raise TypeError("Input must be a pandas DataFrame.")
        # make sure index freq is the same (because pandas has a bug that some
        # operation, e.g. concat, may change freq)
        predicted.index.freq = df.index.freq
        return predicted


class _TrainableMultivariateModel(_TrainableModel):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._cols = []  # type: List[str]

    def _fit(self, df: pd.DataFrame) -> None:
        if isinstance(df, pd.DataFrame):
            if df.columns.duplicated().any():
                raise ValueError(
                    "Input DataFrame must have unique column names."
                )
            df_copy = df.copy()
            self._fit_core(df_copy)
        else:
            raise TypeError("Input must be a pandas DataFrame.")
        self._cols = list(df.columns)
        self._fitted = 1

    def _predict(self, df: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        if self._fitted == 0:
            raise RuntimeError("The model must be trained first.")
        if isinstance(df, pd.DataFrame):
            if df.columns.duplicated().any():
                raise ValueError(
                    "Input DataFrame must have unique column names."
                )
            if not (set(df.columns) >= set(self._cols)):
                raise ValueError(
                    "The model was trained by a pandas DataFrame with columns "
                    "{}, but the input DataFrame does not contain columns {}.".format(
                        self._cols, list(set(self._cols) - set(df.columns))
                    )
                )
            df_copy = (
                df.loc[:, self._cols].copy() if self._cols else df.copy()
            )  # in a customized hd model that doesn't need fit, self._cols is empty
            predicted = self._predict_core(df_copy)
        else:
            raise TypeError("Input must be a pandas DataFrame.")
        # make sure index freq is the same (because pandas has a bug that some
        # operation, e.g. concat, may change freq)
        predicted.index.freq = df.index.freq
        return predicted



"""Module for high-dimensional transformers.

High-dimensional transformers transform hight-dimensional time series, i.e.
pandas DataFrame, into different series, to extract useful information out of
the original time series.

"""

from typing import Any, Callable, Dict, Optional, Tuple, Union

import pandas as pd
from sklearn.decomposition import PCA

from .._transformer_base import (
    _NonTrainableMultivariateTransformer,
    _TrainableMultivariateTransformer,
)


class CustomizedTransformerHD(_TrainableMultivariateTransformer):
    """Multivariate transformer derived from a user-given function and parameters.

    Parameters
    ----------
    Parameters
    ----------
    transform_func: function
        A function transforming multivariate time series.

        The first input argument must be a pandas DataFrame, optional input
        argument may be accepted through parameter `transform_func_params` and
        the output of `fit_func`, and the output must be a pandas Series or
        DataFrame with the same index as input.

    transform_func_params: dict, optional
        Parameters of `transform_func`. Default: None.

    fit_func: function, optional
        A function training parameters of `transform_func` with multivariate
        time series.

        The first input argument must be a pandas DataFrame, optional input
        argument may be accepted through parameter `fit_func_params`, and the
        output must be a dict that can be used by `transform_func` as
        parameters. Default: None.

    fit_func_params: dict, optional
        Parameters of `fit_func`. Default: None.

    """

    def __init__(
        self,
        transform_func: Callable,
        transform_func_params: Optional[Dict[str, Any]] = None,
        fit_func: Optional[Callable] = None,
        fit_func_params: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._fitted_transform_func_params = {}  # type: Dict
        super().__init__()
        self.transform_func = transform_func
        self.transform_func_params = transform_func_params
        self.fit_func = fit_func
        self.fit_func_params = fit_func_params
        if self.fit_func is None:
            self._fitted = 1

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return (
            "transform_func",
            "transform_func_params",
            "fit_func",
            "fit_func_params",
        )

    def _fit_core(self, df: pd.DataFrame) -> None:
        if self.fit_func is not None:
            if self.fit_func_params is not None:
                fit_func_params = self.fit_func_params
            else:
                fit_func_params = {}
            self._fitted_transform_func_params = self.fit_func(
                df, **fit_func_params
            )

    def _predict_core(
        self, df: pd.DataFrame
    ) -> Union[pd.Series, pd.DataFrame]:
        if self.transform_func_params is not None:
            transform_func_params = self.transform_func_params
        else:
            transform_func_params = {}
        if self.fit_func is not None:
            return self.transform_func(
                df,
                **{
                    **self._fitted_transform_func_params,
                    **transform_func_params,
                }
            )
        else:
            return self.transform_func(df, **transform_func_params)


class SumAll(_NonTrainableMultivariateTransformer):
    """Transformer that returns the sum all series as one series."""

    def __init__(self) -> None:
        super().__init__()

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return tuple()

    def _predict_core(self, df: pd.DataFrame) -> pd.Series:
        return df.sum(axis=1, skipna=False)


class RegressionResidual(_TrainableMultivariateTransformer):
    """Transformer that performs regression to build relationship between a
    target series and the rest of series, and returns regression residual
    series.

    Parameters
    ----------
    regressor: object
        Regressor to be used. Same as a scikit-learn regressor, it should
        minimally have `fit` and `predict` methods.
    target: str, optional
        Name of the column to be regarded as target variable.

    """

    def __init__(self, regressor: Any, target: str) -> None:
        super().__init__()
        self.regressor = regressor
        self.target = target

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("regressor", "target")

    def _fit_core(self, df: pd.DataFrame) -> None:
        if self.target not in df.columns:
            raise RuntimeError(
                "Cannot find target series {} in input dataframe.".format(
                    self.target
                )
            )
        self._target = self.target
        self._features = [col for col in df.columns if col != self._target]
        if df.dropna().empty:
            raise RuntimeError("Valid values are not enough for training.")
        self.regressor.fit(
            df.dropna().loc[:, self._features],
            df.dropna().loc[:, self._target],
        )

    def _predict_core(self, df: pd.DataFrame) -> pd.Series:
        target = self._target
        features = self._features
        if target not in df.columns:
            raise RuntimeError(
                "Cannot find target series {} in input dataframe.".format(
                    target
                )
            )
        if not set(features) <= set(df.columns):
            raise RuntimeError(
                "The following series are not found in input dataframe: {}.".format(
                    set(features) - set(df.columns)
                )
            )
        residual = pd.Series(index=df.index, dtype="float64")
        if not df.dropna().empty:
            residual.loc[df.dropna().index] = df.dropna().loc[
                :, target
            ] - self.regressor.predict(df.dropna().loc[:, features])
        return residual


class PcaProjection(_TrainableMultivariateTransformer):
    """Transformer that performs principal component analysis (PCA) to the
    multivariate time series (every time point is treated as a point in high-
    dimensional space), and represents those points with their projection on
    the first k principal components.

    Parameters
    ----------
    k: int, optional
        Number of principal components to use. Default: 1.

    """

    def __init__(self, k: int = 1) -> None:
        self._model = None  # type: Any
        super().__init__()
        self.k = k

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("k",)

    def _fit_core(self, df: pd.DataFrame) -> None:
        self._model = PCA(n_components=self.k)
        if df.dropna().empty:
            raise RuntimeError("Valid values are not enough for training.")
        self._model.fit(df.dropna().values)

    def _predict_core(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.k > self._model.n_components:
            raise ValueError(
                "k is increased after previous fitting. Please fit again."
            )
        results = pd.DataFrame(
            index=df.index, columns=["pc{}".format(i) for i in range(self.k)]
        )
        if not df.dropna().empty:
            results.loc[df.dropna().index] = self._model.transform(
                df.dropna().values
            )[:, : self.k]
        return results


class PcaReconstruction(_TrainableMultivariateTransformer):
    """Transformer that performs principal component analysis (PCA) to the
    multivariate time series  (every time point is treated as a point in high-
    dimensional space), and reconstructs those points with the first k
    principal components.

    Parameters
    ----------
    k: int, optional
        Number of principal components to use. Default: 1.

    """

    def __init__(self, k: int = 1) -> None:
        self._model = None  # type: Any
        super().__init__()
        self.k = k

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("k",)

    def _fit_core(self, df: pd.DataFrame) -> None:
        self._model = PCA(n_components=self.k)
        if df.dropna().empty:
            raise RuntimeError("Valid values are not enough for training.")
        self._model.fit(df.dropna().values)

    def _predict_core(self, df: pd.DataFrame) -> pd.DataFrame:
        if self._model is None:
            raise RuntimeError("Please fit the model first.")
        if self.k > self._model.n_components:
            raise ValueError(
                "k is increased after previous fitting. Please fit again."
            )
        results = pd.DataFrame(columns=df.columns, index=df.index)
        if not df.dropna().empty:
            results.loc[df.dropna().index] = self._model.inverse_transform(
                self._model.transform(df.dropna().values)
            )
        return results


class PcaReconstructionError(_TrainableMultivariateTransformer):
    """Transformer that performs principal component analysis (PCA) to the
    multivariate time series  (every time point is treated as a point in high-
    dimensional space), reconstruct those points with the first k principal
    components, and returns the reconstruction error (i.e. squared distance
    bewteen the reconstructed point and original point).

    Parameters
    ----------
    k: int, optional
        Number of principal components to use. Default: 1.

    """

    def __init__(self, k: int = 1) -> None:
        self._model = None  # type: Any
        super().__init__()
        self.k = k

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("k",)

    def _fit_core(self, df: pd.DataFrame) -> None:
        self._model = PCA(n_components=self.k)
        if df.dropna().empty:
            raise RuntimeError("Valid values are not enough for training.")
        self._model.fit(df.dropna().values)

    def _predict_core(self, df: pd.DataFrame) -> pd.Series:
        if self._model is None:
            raise RuntimeError("Please fit the model first.")
        if self.k > self._model.n_components:
            raise ValueError(
                "k is increased after previous fitting. Please fit again."
            )
        results = pd.DataFrame(columns=df.columns, index=df.index)
        if not df.dropna().empty:
            results.loc[df.dropna().index] = self._model.inverse_transform(
                self._model.transform(df.dropna().values)
            )
        return ((results - df) ** 2).sum(axis=1, skipna=False)



from typing import Any, Callable, Dict, List, Tuple, Union

import pandas as pd

from ._base import (
    _NonTrainableUnivariateModel,
    _TrainableMultivariateModel,
    _TrainableUnivariateModel,
)
from .data import to_events
from .metrics import f1_score, iou, precision, recall


class _NonTrainableUnivariateDetector(_NonTrainableUnivariateModel):
    def predict(
        self, ts: Union[pd.Series, pd.DataFrame], return_list: bool = False
    ) -> Union[
        pd.Series,
        pd.DataFrame,
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
    ]:
        """Detect anomalies from given time series.

        Parameters
        ----------
        ts: pandas.Series or pandas.DataFrame
            Time series to detect anomalies from. If a DataFrame with k
            columns, it is treated as k independent univariate time series, and
            the detector will be applied to each univariate series
            independently.

        return_list: bool, optional
            Whether to return a list of anomalous events, or a binary series
            indicating normal/anomalous. Default: False.

        Returns
        -------
        pandas.Series, pandas.DataFrame, list, or dict
            Detected anomalies.

            - If input is a Series and return_list=False, return a Series;
            - If input is a DataFrame and return_list=False, return a
              DataFrame, where each column corresponds a column in input;
            - If input is a Series and return_list=True, return a list of time
              stamps or time stamp tuples;
            - If input is a DataFrame and return_list=True, return a dict of
              lists, where each key-value pair corresponds a column in input.

        """
        detected = self._predict(ts)
        if return_list:
            return to_events(detected)
        else:
            return detected

    detect = predict

    def score(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        anomaly_true: Union[
            pd.Series,
            pd.DataFrame,
            List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            Dict[
                str,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            ],
        ],
        scoring: str = "recall",
        **kwargs: Any
    ) -> Union[float, Dict[str, float]]:
        """Detect anomalies and score the results against true anomalies.

        Parameters
        ----------
        ts: pandas Series or pandas.DataFrame
            Time series to detect anomalies from.
            If a DataFrame with k columns, it is treated as k independent
            univariate time series, and the detector will be applied to each
            series independently.

        anomaly_true: pandas.Series, pandas.DataFrame, list, or dict
            True anomalies.

            - If pandas Series, it is treated as a series of binary labels.
            - If pandas DataFrame, each column is a binary series and is
              treated as an independent type of anomaly.
            - If list, a list of events where an event is a pandas Timestamp if
              it is instantaneous or a 2-tuple of pandas Timestamps if it is a
              closed time interval.
            - If dict, each key-value pair is a list of events and is treated
              as an independent type of anomaly.

        scoring: str, optional
            Scoring function to use. Must be one of "recall", "precision",
            "f1", and "iou". See module `metrics` for more information.
            Default: "recall"

        **kwargs
            Optional parameters for scoring function. See module `metrics` for
            more information.

        Returns
        -------
        float or dict
            Score(s) for each type of anomaly.

        """
        if scoring == "recall":
            scoring_func = recall  # type: Callable
        elif scoring == "precision":
            scoring_func = precision
        elif scoring == "f1":
            scoring_func = f1_score
        elif scoring == "iou":
            scoring_func = iou
        else:
            raise ValueError(
                "Argument `scoring` must be one of 'recall', 'precision', "
                "'f1' and 'iou'."
            )
        if isinstance(anomaly_true, (pd.Series, pd.DataFrame)):
            return scoring_func(
                y_true=anomaly_true,
                y_pred=self.detect(ts, return_list=False),
                **kwargs
            )
        else:
            return scoring_func(
                y_true=anomaly_true,
                y_pred=self.detect(ts, return_list=True),
                **kwargs
            )


class _TrainableUnivariateDetector(_TrainableUnivariateModel):
    def fit(self, ts: Union[pd.Series, pd.DataFrame]) -> None:
        """Train the detector with given time series.

        Parameters
        ----------
        ts: pandas.Series or pandas.DataFrame
            Time series to be used to train the detector.
            If a DataFrame with k columns, k univariate detectors will be
            trained independently.

        """
        self._fit(ts)

    def predict(
        self, ts: Union[pd.Series, pd.DataFrame], return_list: bool = False
    ) -> Union[
        pd.Series,
        pd.DataFrame,
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
    ]:
        """Detect anomalies from given time series.

        Parameters
        ----------
        ts: pandas.Series or pandas.DataFrame
            Time series to detect anomalies from. If a DataFrame with k
            columns, it is treated as k independent univariate time series.

            - If the detector was trained with a Series, the detector will be
              applied to each univariate series independently;
            - If the detector was trained with a DataFrame, i.e. the detector
              is essentially k detectors, those detectors will be applied to
              each univariate series respectively.

        return_list: bool, optional
            Whether to return a list of anomalous events, or a binary series
            indicating normal/anomalous. Default: False.

        Returns
        -------
        pandas.Series, pandas.DataFrame, list, or dict
            Detected anomalies.

            - If input is a Series and return_list=False, return a Series;
            - If input is a DataFrame and return_list=False, return a
              DataFrame, where each column corresponds a column in input;
            - If input is a Series and return_list=True, return a list of
              events where an event is a pandas Timestamp if it is
              instantaneous or a 2-tuple of pandas Timestamps if it is a closed
              time interval.
            - If input is a DataFrame and return_list=True, return a dict of
              event lists, where each key-value pair corresponds a column in
              input.

        """
        detected = self._predict(ts)
        if return_list:
            return to_events(detected)
        else:
            return detected

    def fit_predict(
        self, ts: Union[pd.Series, pd.DataFrame], return_list: bool = False
    ) -> Union[
        pd.Series,
        pd.DataFrame,
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
    ]:
        """Train the detector and detect anomalies from the time series used
        for training.

        Parameters
        ----------
        ts: pandas.Series or pandas.DataFrame
            Time series to be used for training and be detected for anomalies.
            If a DataFrame with k columns, it is treated as k independent
            univariate time series, and k univariate detectors will be trained
            and applied to each series independently.

        return_list: bool, optional
            Whether to return a list of anomalous events, or a binary series
            indicating normal/anomalous. Default: False.

        Returns
        -------
        pandas.Series, pandas.DataFrame, list, or dict
            Detected anomalies.

            - If input is a Series and return_list=False, return a Series;
            - If input is a DataFrame and return_list=False, return a
              DataFrame, where each column corresponds a column in input;
            - If input is a Series and return_list=True, return a list of
              events where an event is a pandas Timestamp if it is
              instantaneous or a 2-tuple of pandas Timestamps if it is a closed
              time interval.
            - If input is a DataFrame and return_list=True, return a dict of
              event lists, where each key-value pair corresponds a column in
              input.

        """
        self.fit(ts)
        return self.detect(ts, return_list=return_list)

    detect = predict
    fit_detect = fit_predict

    def score(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        anomaly_true: Union[
            pd.Series,
            pd.DataFrame,
            List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            Dict[
                str,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            ],
        ],
        scoring: str = "recall",
        **kwargs: Any
    ) -> Union[float, Dict[str, float]]:
        """Detect anomalies and score the results against true anomalies.

        Parameters
        ----------
        ts: pandas Series or pandas.DataFrame
            Time series to detect anomalies from.
            If a DataFrame with k columns, it is treated as k independent
            univariate time series, and k univariate detectors will be applied
            to each series independently.

        anomaly_true: pandas.Series, pandas.DataFrame, list, or dict
            True anomalies.

            - If pandas Series, it is treated as a series of binary labels.
            - If pandas DataFrame, each column is a binary series and is
              treated as an independent type of anomaly.
            - If list, a list of events where an event is a pandas Timestamp if
              it is instantaneous or a 2-tuple of pandas Timestamps if it is a
              closed time interval.
            - If dict, each key-value pair is a list of events and is treated
              as an independent type of anomaly.

        scoring: str, optional
            Scoring function to use. Must be one of "recall", "precision",
            "f1", and "iou". See module `metrics` for more information.
            Default: "recall"

        **kwargs
            Optional parameters for scoring function. See module `metrics` for
            more information.

        Returns
        -------
        float or dict
            Score(s) for each type of anomaly.

        """
        if scoring == "recall":
            scoring_func = recall  # type: Callable
        elif scoring == "precision":
            scoring_func = precision
        elif scoring == "f1":
            scoring_func = f1_score
        elif scoring == "iou":
            scoring_func = iou
        else:
            raise ValueError(
                "Argument `scoring` must be one of 'recall', 'precision', "
                "'f1' and 'iou'."
            )
        if isinstance(anomaly_true, (pd.Series, pd.DataFrame)):
            return scoring_func(
                y_true=anomaly_true,
                y_pred=self.detect(ts, return_list=False),
                **kwargs
            )
        else:
            return scoring_func(
                y_true=anomaly_true,
                y_pred=self.detect(ts, return_list=True),
                **kwargs
            )


# class _NonTrainableMultivariateDetector(_NonTrainableMultivariateModel):
#     def detect(
#         self, df: pd.DataFrame, return_list: bool = False
#     ) -> Union[
#         pd.Series, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
#     ]:
#         """Detect anomalies from given time series.

#         Parameters
#         ----------
#         df: pandas.DataFrame
#             Time series to detect anomalies from.

#         return_list: bool, optional
#             Whether to return a list of anomalous time stamps, or a binary
#             series indicating normal/anomalous. Default: False.

#         Returns
#         -------
#         pandas.Series or list
#             Detected anomalies.

#             - If return_list=False, return a binary series;
#             - If return_list=True, return a list of time stamps or time stamp
#               2-tuples.

#         """
#         detected = self._predict(df)
#         if return_list:
#             return to_events(detected)
#         else:
#             return detected

#     def predict(
#         self, df: pd.DataFrame, return_list: bool = False
#     ) -> Union[
#         pd.Series, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
#     ]:
#         """
#         Alias of `detect`.
#         """
#         return self.detect(df, return_list=return_list)

#     def score(
#         self,
#         df: pd.DataFrame,
#         anomaly_true: Union[pd.Series, List, Tuple],
#         scoring: str = "recall",
#         **kwargs: Any
#     ) -> float:
#         """Detect anomalies and score the results against true anomalies.

#         Parameters
#         ----------
#         df: pandas DataFrame
#             Time series to detect anomalies from.
#             If a DataFrame with k columns, k univariate detectors will be
#             applied to them independently.

#         anomaly_true: Series, or a list of Timestamps or Timestamp tuple
#             True anomalies.

#             - If Series, it is a series binary labels indicating anomalous;
#             - If list, it is a list of anomalous events in form of time windows.

#         scoring: str, optional
#             Scoring function to use. Must be one of "recall", "precision",
#             "f1", and "iou". See module `metrics` for more information.
#             Default: "recall"

#         **kwargs
#             Optional parameters for scoring function. See module `metrics` for
#             more information.

#         Returns
#         -------
#         float
#             Score of detection result.

#         """
#         if scoring == "recall":
#             scoring_func = recall  # type: Callable
#         elif scoring == "precision":
#             scoring_func = precision
#         elif scoring == "f1":
#             scoring_func = f1_score
#         elif scoring == "iou":
#             scoring_func = iou
#         else:
#             raise ValueError(
#                 "Argument `scoring` must be one of 'recall', 'precision', "
#                 "'f1' and 'iou'."
#             )
#         if isinstance(anomaly_true, pd.Series):
#             return scoring_func(
#                 y_true=anomaly_true,
#                 y_pred=self.detect(df, return_list=False),
#                 **kwargs
#             )
#         else:
#             return scoring_func(
#                 y_true=anomaly_true,
#                 y_pred=self.detect(df, return_list=True),
#                 **kwargs
#             )


class _TrainableMultivariateDetector(_TrainableMultivariateModel):
    def fit(self, df: pd.DataFrame) -> None:
        """Train the detector with given time series.

        Parameters
        ----------
        df: pandas.DataFrame
            Time series to be used to train the detector.

        """
        self._fit(df)

    def predict(
        self, df: pd.DataFrame, return_list: bool = False
    ) -> Union[
        pd.Series, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ]:
        """Detect anomalies from given time series.

        Parameters
        ----------
        df: pandas.DataFrame
            Time series to detect anomalies from.

        return_list: bool, optional
            Whether to return a list of anomalous events, or a binary series
            indicating normal/anomalous. Default: False.

        Returns
        -------
        pandas.Series or list
            Detected anomalies.

            - If return_list=False, return a binary series;
            - If return_list=True, return a list of events where an event is a
              pandas Timestamp if it is instantaneous or a 2-tuple of pandas
              Timestamps if it is a closed time interval.

        """
        detected = self._predict(df)
        if return_list:
            return to_events(detected)
        else:
            return detected

    def fit_predict(
        self, df: pd.DataFrame, return_list: bool = False
    ) -> Union[
        pd.Series, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ]:
        """Train the detector and detect anomalies from the time series used
        for training.

        Parameters
        ----------
        df: pandas.DataFrame
            Time series to be used for training and be detected for anomalies.

        return_list: bool, optional
            Whether to return a list of anomalous events, or a binary series
            indicating normal/anomalous. Default: False.

        Returns
        -------
        pandas.Series or list
            Detected anomalies.

            - If return_list=False, return a binary series;
            - If return_list=True, return a list of events where an event is a
              pandas Timestamp if it is instantaneous or a 2-tuple of pandas
              Timestamps if it is a closed time interval.

        """
        self.fit(df)
        return self.detect(df, return_list=return_list)

    detect = predict
    fit_detect = fit_predict

    def score(
        self,
        df: pd.DataFrame,
        anomaly_true: Union[pd.Series, List, Tuple],
        scoring: str = "recall",
        **kwargs: Any
    ) -> float:
        """Detect anomalies and score the results against true anomalies.

        Parameters
        ----------
        df: pandas DataFrame
            Time series to detect anomalies from.
            If a DataFrame with k columns, k univariate detectors will be
            applied to them independently.

        anomaly_true: Series or list
            True anomalies.

            - If pandas Series, it is treated as a series of binary labels.
            - If list, a list of events where an event is a pandas Timestamp if
              it is instantaneous or a 2-tuple of pandas Timestamps if it is a
              closed time interval.

        scoring: str, optional
            Scoring function to use. Must be one of "recall", "precision",
            "f1", and "iou". See module `metrics` for more information.
            Default: "recall"

        **kwargs
            Optional parameters for scoring function. See module `metrics` for
            more information.

        Returns
        -------
        float
            Score of detection result.

        """
        if scoring == "recall":
            scoring_func = recall  # type: Callable
        elif scoring == "precision":
            scoring_func = precision
        elif scoring == "f1":
            scoring_func = f1_score
        elif scoring == "iou":
            scoring_func = iou
        else:
            raise ValueError(
                "Argument `scoring` must be one of 'recall', 'precision', "
                "'f1' and 'iou'."
            )
        if isinstance(anomaly_true, pd.Series):
            return scoring_func(
                y_true=anomaly_true,
                y_pred=self.detect(df, return_list=False),
                **kwargs
            )
        else:
            return scoring_func(
                y_true=anomaly_true,
                y_pred=self.detect(df, return_list=True),
                **kwargs
            )
